/*
 * $Id: cilk-conf.h,v 1.32 1996/06/10 15:46:23 athena Exp $
 */
/*
 * $Log: cilk-conf.h,v $
 * Revision 1.32  1996/06/10 15:46:23  athena
 * CRITICAL_PATH is on by default in the rts, but off in user
 * programs. This is a hack! In this way, the user doesn't need to
 * recompile the rts, just her program.
 *
 * Revision 1.31  1996/06/07 17:55:43  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.30  1996/06/06 16:06:34  athena
 * Fixed statistics code.
 *
 * Revision 1.29  1996/06/04 21:33:21  athena
 * Added READMEs and some comments
 *
 * Revision 1.28  1996/05/30 21:43:24  athena
 * Minor changes. We now use a static message for assertions, in
 * order to save space.
 *
 * Revision 1.27  1996/05/30 20:10:32  athena
 * Added copyright legalese
 *
 * Revision 1.26  1996/05/29 22:13:09  athena
 * Added trace statements for abort
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_CONF_H
#define _CILK_CONF_H

/*
 * Keep track of various statistics
 */
#ifndef CILK_STATS
#define CILK_STATS 1
#endif

/*
 * Keep track of execution time
 */
#ifndef CILK_TIMING
#define CILK_TIMING 1
#endif

/*
 * Measure critical path. VERY EXPENSIVE in terms of execution time
 */
#ifndef CILK_CRITICAL_PATH
#define CILK_CRITICAL_PATH 1
#endif

/*
 * Enable internal rts tracing
 */
#ifndef CILK_TRACE
#define CILK_TRACE 1
#endif

/*
 * Enable rts self-consistency checks.
 */
#ifndef CILK_DEBUG
#define CILK_DEBUG 1
#endif

#define MEGABYTE (1UL<<20)

/* 
 * Default size of Cilk's stack
 */
#define CILK_DEFAULT_STACKSIZE (512U * MEGABYTE)

/* 
 * Default size of Cilk's heap
 */
#define CILK_DEFAULT_HEAPSIZE (64U * MEGABYTE)

/*
 * Note: We can get just under 2GB of virtual memory space for our own
 * use under Solaris.  These defaults use less than 1GB, so there is some
 * room to spare if you need it.  Your operating system may vary.
 */
#define CILK_DEFAULT_BSTORESIZE (64U * MEGABYTE)

/*
 *  Architecture-specific stuff.
 */

#ifndef CILK_SOLARIS
#define CILK_SOLARIS 0
#endif

#ifndef CILK_IRIX
#define CILK_IRIX 0
#endif

#if CILK_SOLARIS
#define CILK_ARCH_H <arch-solaris.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define LOG_CACHE_LINE 6
#endif

#if CILK_IRIX
#define CILK_ARCH_H <arch-irix.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define LOG_CACHE_LINE 6
#endif

#ifndef CILK_ARCH_H
#error No architecture defined
#endif

/*
 * there should be no need to hack below this line
 */
#if CILK_CRITICAL_PATH && (!CILK_TIMING)
#warning Must #define CILK_TIMING when CILK_CRITICAL_PATH is 1
#undef CILK_TIMING
#define CILK_TIMING 1
#endif

#define CILK_CHECK(cond, complaints) \
     if (!(cond)) Cilk_die complaints;

#define CILK_DPRINTF Cilk_dprintf

#define CILK_FRAME_STEALABLE_MAGIC (0xDEADBEEF)
#define CILK_FRAME_PROVSTEAL_MAGIC (0xDEBEEFED)
#define CILK_FRAME_NONSTEALABLE_MAGIC (0xDEADFEEB)
#define CILK_FRAME_STOLEN_MAGIC (0xFEEBDEAD)

#if CILK_DEBUG
#define CILK_ASSERT(ex) \
      (void)((ex) || (Cilk_die(Cilk_assertion_failed_msg, \
             #ex, __LINE__, __FILE__), 0))
#define WHEN_CILK_DEBUG(ex)  ex
#else
#define CILK_ASSERT(ex)
#define WHEN_CILK_DEBUG(ex)
#endif

#if CILK_TRACE

#define TRACE(mode,a) if ((mode) & Cilk_trace_mode) CILK_DPRINTF a
#define WHEN_CILK_TRACE(a) a
/* Trace modes */
#define TRACE_NONE 0x0001	/* reserved */
#define TRACE_SETUP 0x0002
#define TRACE_BSTORE 0x0004
#define TRACE_PAGE_TABLE 0x0008
#define TRACE_SIGNAL 0x0010
#define TRACE_RETURN_SUBCOMP 0x0020
#define TRACE_STEAL 0x0040
#define TRACE_MISC 0x0080
#define TRACE_MAIN 0x0100
#define TRACE_MALLOC 0x0200
#define TRACE_ABORT 0x0400

#else				/* CILK_TRACE */

#define TRACE(mode,a)
#define WHEN_CILK_TRACE(a)

#endif				/* CILK_TRACE */

#if CILK_STATS
#define WHEN_CILK_STATS(a) a
#else
#define WHEN_CILK_STATS(a)
#endif

#if CILK_TIMING
#define WHEN_CILK_TIMING(a) a
#else
#define WHEN_CILK_TIMING(a)
#endif

#if CILK_CRITICAL_PATH
#define WHEN_CILK_CP(a) a
#else
#define WHEN_CILK_CP(a)
#endif

#define CACHE_LINE (1<<LOG_CACHE_LINE)
#define NEXT_CACHE_LINE(p) ((void*)(((size_t)(p)+CACHE_LINE-1) & ~(CACHE_LINE-1)))

#endif				/* _CILK_CONF_H */
