/*
 * $Id: cilk.c,v 1.153 1996/06/10 15:46:26 athena Exp $
 */

/*
 * $Log: cilk.c,v $
 * Revision 1.153  1996/06/10 15:46:26  athena
 * CRITICAL_PATH is on by default in the rts, but off in user
 * programs. This is a hack! In this way, the user doesn't need to
 * recompile the rts, just her program.
 *
 * Revision 1.152  1996/06/07 17:55:48  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.151  1996/06/06 23:06:57  athena
 * Added printout of average parallelism
 *
 * Revision 1.150  1996/06/06 16:06:37  athena
 * Fixed statistics code.
 *
 * Revision 1.149  1996/06/06 14:37:41  athena
 * cilk_main is now main
 *
 * Revision 1.148  1996/06/06 00:18:09  athena
 * Added printout of total execution time.
 *
 * Revision 1.147  1996/06/05 15:37:45  athena
 * Added cilk-lib.h, and updated stuff
 *
 * Revision 1.146  1996/06/04 16:05:50  athena
 * Added cilk-compat.h
 *
 * Revision 1.145  1996/06/03 21:20:45  athena
 * Minor cleanup
 *
 * Revision 1.144  1996/06/03 16:49:37  athena
 * Restructuring of returnit() and steal(), some polishing of Don's
 * checkers code.
 *
 * Revision 1.143  1996/05/31 16:16:42  randall
 * Bug fix for inlet argument copying.
 *
 * Revision 1.142  1996/05/31 02:40:30  randall
 * Fixed THE bug.  Hopefully everything is OK now.
 *
 * Revision 1.141  1996/05/31 02:23:41  randall
 * More bug fixing.  Make frame pointer NULL for aborting subcomputations.
 * Put in checks to make sure inlet arguments are deallocated.
 *
 * Revision 1.140  1996/05/30 22:13:38  randall
 * Updated apply_result to take return_loc structure.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <stdlib.h>
#include <stdio.h>

/***********************************************************\
 * Per-processor Variables
\***********************************************************/

/*
 * make sure everything in here has an initialization!  Otherwise
 * the linker reorders and relocates everything at will.  Also,
 * make sure nothing is declared as const or static, as these
 * variables may go in a different region of memory as well.
 */

/* first data segment variable */
char Cilk_rts_data_begin[CILK_PAGESIZE_HACK] ={1};

char Cilk_result[1024] ={1};	/* TODO */

/* user visible */
int Self = -1;

/* cilk2c visible */
int Cilk_result_valid = 0;
int Cilk_resume = 0;
int Cilk_abort = 0;
volatile StealStuffT *Cilk_ss = NULL;
WHEN_CILK_CP(TimeT Cilk_cp_hack = 0;)

/* rts only */
volatile int Cilk_allow_faults = 0;
volatile int Cilk_last_mapped_page = -1;
u_long Cilk_rand_next = 0;

/* end of per-processor variables */
char Cilk_rts_data_mid[CILK_PAGESIZE_HACK] = {1};

#define INFOFILE stderr		/* file where info messages are sent */

/***********************************************************\
 * Random library functions for use by user
\***********************************************************/

static u_long rand_nxt = 0;
static AlignedLockT rand_lock;

int Cilk_rand(void)
{
     int result;
     Cilk_sema_wait(&rand_lock.lock);
     result = ((rand_nxt = rand_nxt * 1103515245 + 12345)
	       % ((u_int) RAND_MAX + 1));
     Cilk_sema_signal(&rand_lock.lock);
     return result;
}

void Cilk_srand(u_int seed)
{
     Cilk_sema_wait(&rand_lock.lock);
     rand_nxt = seed;
     Cilk_sema_signal(&rand_lock.lock);
}

static void lib_function_init()
{
     Cilk_sema_init(&rand_lock.lock);
}

/***********************************************************\
 * Random Junk (shared variables from now on)
\***********************************************************/

const char *Cilk_assertion_failed_msg = 
     "Assertion failed: %s line %d file %s\n"
     /* Sorry, couldn't resist --- athena */
     "Ex absurdo sequitur quodlibet.\n";

static volatile int done;	/* flag set when cilk_main returns */
static volatile int retval = 0;	/* value that cilk_main returned */

static int CilkStatLevel = 1;
long Cilk_trace_mode = 0;
static size_t arenasize = CILK_DEFAULT_HEAPSIZE;

static size_t pagesize;		/* size of an OS page */
static TimeT start_time;	/* start of cilk program */
int Cilk_active_size = 1;	/* number of processes */

static void *frame_stack;	/* area of memory for frame stack */

/* proc_mem: size of frame stack */
static size_t proc_mem = CILK_DEFAULT_STACKSIZE;

TimeT Cilk_critical_path;

/***********************************************************\
 * Statistics gathering
\***********************************************************/

#if CILK_STATS

typedef struct {
     CACHE_LINE_ALIGNED(int steal_attempts);
     int steals;
     volatile int page_faults;
} StatsT;

static StatsT *stat_array;

static void stats_gathering_init()
{
     int i;

     stat_array = (StatsT *) Cilk_malloc(sizeof(StatsT) * Cilk_active_size);

     for (i = 0; i < Cilk_active_size; i++) {
	  stat_array[i].steal_attempts = 0;
	  stat_array[i].steals = 0;
	  stat_array[i].page_faults = 0;
     }
}

static void print_rts_statistics(void)
{
     int i;

     fprintf(INFOFILE,
	     "\nRUNTIME SYSTEM STATISTICS:\n"
	     "\n"
	     " PN   Attempts     Steals     Faults\n"
	     "------------------------------------\n");

     for (i = 0; i < Cilk_active_size; i++)
	  fprintf(INFOFILE, "%3d%11d%11d%11d\n", i,
		  stat_array[i].steal_attempts,
		  stat_array[i].steals,
		  stat_array[i].page_faults);

     fprintf(INFOFILE, "\n");
}

#define CILK_STAT_STEAL_ATTEMPT() stat_array[Self].steal_attempts++;
#define CILK_STAT_STEAL() stat_array[Self].steals++;
#define CILK_STAT_PAGE_FAULT() stat_array[Self].page_faults++;

#else

#define CILK_STAT_STEAL_ATTEMPT()
#define CILK_STAT_STEAL()
#define CILK_STAT_PAGE_FAULT()

#endif				/* CILK_STATS */


/***********************************************************
 * Time-gathering stuff
 ***********************************************************/

#if CILK_TIMING

/*
 * The model for time gathering is the following:
 * there is a certain set of states in which the system can
 * be. Throughout the code, ENTER_STATE(something) signals
 * that a given state is entered. EXIT_STATE(state) reverts back to
 * the previous state, and checks that the right state is being
 * exited. The time is accumulated on a per-state
 * basis. Time is cumulative, i.e. it is measured from ENTER_STATE
 * to EXIT_STATE, regardless of other states entered in between.
 * 
 * No more than STATE_DEPTH states can be nested.
 * The stack is really used only for checking purposes, but it's
 * probably a good idea to keep it.
 *
 */
#define STATE_DEPTH 256

enum {
     STATE_TOTAL,
     STATE_COPYING_PAGE_TABLE,
     STATE_FREEING_PAGE_TABLE,
     STATE_STEALING,
     STATE_GET_PAGE_ENTRY,
     STATE_GET_BSTORE_PAGE,
     STATE_FREE_BSTORE_PAGE,
     STATE_WAIT_BS_LOCK,
     STATE_ALLOC_SHMEM,
     STATE_FREE_SHMEM,
     STATE_MAPPING_PAGE,
     STATE_WORKING,
     TIME_NSTATES		/* this must be last */
};

/* also update name table below */

typedef struct {
     CACHE_LINE_ALIGNED(volatile int state_stackp);
     volatile int state_stack[STATE_DEPTH];
     volatile TimeT state_stack_begintime[STATE_DEPTH];
     volatile TimeT state_stack_begintime_wall[STATE_DEPTH];
     volatile TimeT statetime[TIME_NSTATES];
     volatile TimeT statetime_wall[TIME_NSTATES];
     volatile int stateentered[TIME_NSTATES];
} StateInfoT;

static StateInfoT *timer_state_info = NULL;

/*
 * table of description of states. It is printed in
 * the same order.
 */
static struct state_desc {
     char *name;
     int state;
} state_descriptions[] = {

     {
	  "Total running time", STATE_TOTAL
     },
     {
	  "Copying page table", STATE_COPYING_PAGE_TABLE
     },
     {
	  "Freeing page table", STATE_FREEING_PAGE_TABLE
     },
     {
	  "Stealing", STATE_STEALING
     },
     {
	  "Getting page table entry", STATE_GET_PAGE_ENTRY
     },
     {
	  "Getting bstore page", STATE_GET_BSTORE_PAGE
     },
     {
	  "Freeing bstore page", STATE_FREE_BSTORE_PAGE
     },
     {
	  "Waiting for bs lock", STATE_WAIT_BS_LOCK
     },
     {
	  "Allocating shmem", STATE_ALLOC_SHMEM
     },
     {
	  "Freeing shmem", STATE_FREE_SHMEM
     },
     {
	  "Mapping shmem pages", STATE_MAPPING_PAGE
     },
     {
	  "Doing useful work", STATE_WORKING
     },
     {
	  (char *) 0, 0
     }
};

static inline void ENTER_STATE(int state)
{
     int stackp;
     TimeT now = Cilk_get_process_time();
     TimeT now_wall = Cilk_get_wall_time();
     StateInfoT *info = timer_state_info + Self;
     if (timer_state_info == NULL)
	  return;
     stackp = info->state_stackp;
     CILK_ASSERT(stackp < STATE_DEPTH);
     info->state_stack[stackp] = state;
     info->state_stack_begintime[stackp] = now;
     info->state_stack_begintime_wall[stackp] = now_wall;
     info->state_stackp = stackp + 1;
     ++(info->stateentered[state]);
}

static inline void EXIT_STATE(int state)
{
     int stackp;
     TimeT now = Cilk_get_process_time();
     TimeT now_wall = Cilk_get_wall_time();
     StateInfoT *info = timer_state_info + Self;
     if (timer_state_info == NULL)
	  return;
     stackp = info->state_stackp - 1;
     info->state_stackp = stackp;
     CILK_ASSERT(stackp >= 0);

     CILK_ASSERT(info->state_stack[stackp] == state);
     info->statetime[state] += now - info->state_stack_begintime[stackp];
     info->statetime_wall[state] += now_wall -
	 info->state_stack_begintime_wall[stackp];
}

/* print time statistics */
static void print_time_statistics(void)
{
     int i;
     StateInfoT *info = timer_state_info;
     double worktime = 0.0;
     double totaltime = 0.0;

     fprintf(INFOFILE,
	     "\nSUMMARY OF TIME SPENT BY EACH PROCESSOR:\n\n"
	     " PN    Working      Total\n"
	     "-------------------------\n");
     for (i = 0; i < Cilk_active_size; i++) {
	  fprintf(INFOFILE, "%3d%11f%11f\n", i,
		  Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]),
		  Cilk_process_time_to_sec(info[i].statetime[STATE_TOTAL]));
	  worktime += 
	       Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]);
	  totaltime += 
	       Cilk_process_time_to_sec(info[i].statetime[STATE_TOTAL]);
     }
     fprintf(INFOFILE,
	     "-------------------------\n"
	     "TOT%11f%11f\n", worktime, totaltime);
}

static void print_detailed_time_statistics(void)
{
     int i;
     struct state_desc *p;
     StateInfoT *info = timer_state_info;

     fprintf(INFOFILE,
	     "\nDETAILS OF TIME SPENT BY EACH PROCESSOR:\n");

     for (i = 0; i < Cilk_active_size; i++) {
	  fprintf(INFOFILE, "\n--- Time statistics for processor %d ---\n",
		  i);

	  for (p = state_descriptions; p->name; ++p) {
	       fprintf(INFOFILE, "%s: %fs / wall %fs (entered %d times)\n",
		       p->name,
		       Cilk_process_time_to_sec(info[i].statetime[p->state]),
		       Cilk_wall_time_to_sec(info[i].statetime_wall[p->state]),
		       info[i].stateentered[p->state]);
	  }
     }
}

static double compute_work(void)
{
     int i;
     double work = 0.0;
     StateInfoT *info = timer_state_info;

     for (i = 0; i < Cilk_active_size; i++) 
	  work +=  Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]);
     
     return work;
}

static void time_gathering_init(void)
{
     int i, j;
     StateInfoT *info;

     info = (StateInfoT *) Cilk_malloc(sizeof(StateInfoT) * Cilk_active_size);
     timer_state_info = info;

     for (i = 0; i < Cilk_active_size; i++) {
	  for (j = 0; j < TIME_NSTATES; ++j) {
	       info[i].statetime[j] = (TimeT) 0;
	       info[i].statetime_wall[j] = (TimeT) 0;
	       info[i].stateentered[j] = 0;
	  }
	  info[i].state_stackp = 0;
     }
}

#else				/* CILK_TIMING */

#define ENTER_STATE(state)
#define EXIT_STATE(state)

#endif

/*
 * The statistics printer 
 */
static void print_all_statistics(void)
{
     if (CilkStatLevel >= 1) {
	  /* level 1 and above: print wall clock, work and CP */
	  fprintf(INFOFILE,
		  "Wall-clock running time on %d processor%s: %f\n",
		  Cilk_active_size, Cilk_active_size > 1 ? "s" : "",
		  Cilk_wall_time_to_sec(Cilk_get_wall_time() - start_time));
	  WHEN_CILK_TIMING(fprintf(INFOFILE, "Total work = %f\n",
				   compute_work()));
	  WHEN_CILK_CP(if (Cilk_critical_path != 0.0)
		       fprintf(INFOFILE, "Critical path = %f\n",
			       Cilk_process_time_to_sec(Cilk_critical_path)));
#if CILK_TIMING && CILK_CRITICAL_PATH
	  if (Cilk_critical_path != 0.0)
	       fprintf(INFOFILE, "Average parallelism = %f\n",
		       compute_work() / 
		       Cilk_process_time_to_sec(Cilk_critical_path));
#endif
     }

     if (CilkStatLevel >= 2) {
	  /* level 2 and above: print statistics on work stealing etc */
	  WHEN_CILK_STATS(print_rts_statistics());
     }

     if (CilkStatLevel >= 3) {
	  /* level 3 and above: print work done by each processor */
	  WHEN_CILK_TIMING(print_time_statistics());
     }

     if (CilkStatLevel >= 4) {
	  /* level 4 and above: print detailed breakout of time */
	  WHEN_CILK_TIMING(print_detailed_time_statistics());
     }
}

/***********************************************************\
 *  Random number generator, since rand()/random() is not
 *  MT-Safe. This RNG is *distinct* from the user RNG, just 
 *  for safety.
 ***********************************************************/
/* code stolen from NetBSD */

/*
 * Cilk_rand_next is stored in the per-processor global section to
 * make this code MT-Safe.
 */

static int rts_rand(void)
{
     return ((Cilk_rand_next = Cilk_rand_next * 1103515245 + 12345)
	     % ((u_int) RAND_MAX + 1));
}

static void rts_srand(u_int seed)
{
     Cilk_rand_next = seed;
}

/***********************************************************\
 * Shared memory locks
\***********************************************************/
/*
 *      Lock Diagram:
 * To ensure that no deadlocks occur, make sure that all locks are
 * obtained in order (downward) along a path in the following partial
 * order.  Subcomputation locks must be obtained in a downward order
 * in the subcomputation tree.
 *
 *                      subcomp lock
 *                           |
 *                           |
 *                      stealer lock
 *                      /    |
 *                     /     |
 *      page table lock      |
 *           |         \     |
 *           |          \    |
 *        bs_lock        shmem_lock
 */

/***********************************************************\
 * Cactus-stack memory allocator
\***********************************************************/

/*
 * We should pass this function the frame pointer so we don't have to
 * read it from Cilk_ss->tail.
 */
void *Cilk_alloca(size_t size)
{
     StackFrameT *frame = (StackFrameT *) Cilk_ss->tail;
     void *mem = (char *) frame + frame->size;

     /* align size */
     size = ((size + CILK_MAX_ALIGN - 1) & ~(CILK_MAX_ALIGN - 1));

     CILK_ASSERT((frame->size & (CILK_MAX_ALIGN - 1)) == 0);
     CILK_ASSERT((((int) frame) & (CILK_MAX_ALIGN - 1)) == 0);
     CILK_ASSERT(frame->magic == CILK_FRAME_NONSTEALABLE_MAGIC);

     /* allocate space in frame */
     frame->size = (char *) mem + size - (char *) frame;

     CILK_ASSERT((char *) mem + size
		 <= (char *) frame_stack + proc_mem);

     return mem;
}

/***********************************************************\
 * Shared memory heap for use by system
\***********************************************************/

/* 
 * currently, only page tables and subcomputations are allocated
 * using this mechanism.
 */

/* amount to malloc at one time */
/* 
 * Allocate just under a power of two so that Cilk_malloc doesn't
 * waste space.  Matteo, please check this.  -Keith
 */
#define SHMEM_SIZE (255*CACHE_LINE)

static void *shmem_end;		/* end of shared memory heap */
static void *shmem_free;	/*
				 * free pointer for shared memory 
				 * kept cache-line aligned 
				 */
static AlignedLockT shmem_lock;	/* lock on shared-memory allocator */
WHEN_CILK_DEBUG(static int shmem_count;) /* amount of allocated memory */
#define MAX_SMALL_ALLOC 4
static volatile void *volatile free_shmem_array[MAX_SMALL_ALLOC];

/*
 * Set up shared-memory heap.  Call this once before processes are forked.
 */
static void setup_shared_memory(void)
{
     int i;
     void *shmem_area;

     shmem_area = Cilk_malloc(SHMEM_SIZE);

     TRACE(TRACE_SETUP,
	   ("got shared-memory area %x\n", (unsigned int) shmem_area));

     shmem_end = shmem_area + SHMEM_SIZE;
     shmem_free = NEXT_CACHE_LINE(shmem_area);
     Cilk_sema_init(&shmem_lock.lock);

     for (i = 0; i < MAX_SMALL_ALLOC; i++)
	  free_shmem_array[i] = NULL;

     /* count amount of allocated shared memory */
     WHEN_CILK_DEBUG(shmem_count = 0);
}

/*
 * System routine for allocating from the system shared memory heap.
 * Used for subcomputations, and page tables.  Returned blocks are
 * CACHE_LINE aligned.
 */
static volatile void *malloc_shmem(unsigned size)
{
     volatile void *area;

     ENTER_STATE(STATE_ALLOC_SHMEM);
     size = (size + CACHE_LINE - 1) >> LOG_CACHE_LINE;
     CILK_ASSERT(size < MAX_SMALL_ALLOC);
     CILK_ASSERT(size > 0);

     Cilk_sema_wait(&shmem_lock.lock);
     WHEN_CILK_DEBUG(shmem_count += size);

     /* reuse small items */
     if ((area = free_shmem_array[size]) != NULL) {
	  free_shmem_array[size] = *((volatile void **) area);
	  Cilk_sema_signal(&shmem_lock.lock);
	  EXIT_STATE(STATE_ALLOC_SHMEM);
	  return area;
     }
     area = shmem_free;
     shmem_free += (size << LOG_CACHE_LINE);

     if (shmem_free > shmem_end) {
	  /* need to Cilk_malloc some more memory! */
	  void *shmem_area = Cilk_malloc(SHMEM_SIZE);

	  TRACE(TRACE_SETUP,
	     ("got more shared-memory %x\n", (unsigned int) shmem_area));

	  shmem_end = shmem_area + SHMEM_SIZE;
	  shmem_free = NEXT_CACHE_LINE(shmem_area);
	  area = shmem_free;
	  shmem_free += (size << LOG_CACHE_LINE);
     }
     CILK_ASSERT(shmem_free <= shmem_end);

     Cilk_sema_signal(&shmem_lock.lock);
     EXIT_STATE(STATE_ALLOC_SHMEM);
     return area;
}

static void free_shmem(volatile void *area, int size)
{
     ENTER_STATE(STATE_FREE_SHMEM);
     size = (size + CACHE_LINE - 1) >> LOG_CACHE_LINE;
     CILK_ASSERT(size < MAX_SMALL_ALLOC);
     CILK_ASSERT(size > 0);

     Cilk_sema_wait(&shmem_lock.lock);
     WHEN_CILK_DEBUG(shmem_count -= size);
     *((volatile void **) area) = free_shmem_array[size];
     free_shmem_array[size] = area;
     Cilk_sema_signal(&shmem_lock.lock);

     EXIT_STATE(STATE_FREE_SHMEM);
}

/***********************************************************
 * Backing store
 ***********************************************************/
/* bstore_mem: size of backing-store file */
static size_t bstore_mem = CILK_DEFAULT_BSTORESIZE;

static volatile int bs_free_list;
static volatile int *bs_flist_array;
WHEN_CILK_DEBUG(static volatile int bs_used_pages;)
static AlignedLockT bs_lock;

static void setup_bstore(void)
{
     int npages;		/* number of pages */
     int i;

     TRACE(TRACE_SETUP, ("Start setup bstore\n"));

     CILK_CHECK(bstore_mem % pagesize == 0,
		("bstore_mem not multiple of pagesize!\n"));

     Cilk_setup_bstore(bstore_mem);

     npages = bstore_mem / pagesize;

     /* allocate a chunk of shared memory to hold the free list */
     bs_flist_array = Cilk_malloc(npages * sizeof(int));

     /* init the free list to contain all pages */
     for (i = 0; i < npages; ++i) {
	  bs_flist_array[i] = i + 1;
     }

     bs_flist_array[npages - 1] = -1;

     /* init the lock */
     Cilk_sema_init(&bs_lock.lock);

     /* initialize the free list pointer */
     bs_free_list = 0;		/* first free page */

     /* count # of used pages */
     WHEN_CILK_DEBUG(bs_used_pages = 0);

     TRACE(TRACE_SETUP, ("End setup bstore\n"));
}

static int get_bstore_page(void)
{
     int page;

     ENTER_STATE(STATE_GET_BSTORE_PAGE);

     ENTER_STATE(STATE_WAIT_BS_LOCK);
     Cilk_sema_wait(&bs_lock.lock);
     EXIT_STATE(STATE_WAIT_BS_LOCK);

     page = bs_free_list;

     CILK_CHECK(page != -1,
		("out of bstore pages!\n"));

     bs_free_list = bs_flist_array[page];

     WHEN_CILK_DEBUG(bs_used_pages++);

     /* 
      * This check ensures that we don't use the same page twice.
      * The second check is always true, but it's not done if not
      * in debugging mode.
      */
     CILK_ASSERT(bs_flist_array[page] != -2);
     CILK_ASSERT(bs_flist_array[page] = -2);

     Cilk_sema_signal(&bs_lock.lock);

     CILK_ASSERT(page >= 0 && page < bstore_mem / pagesize);
     EXIT_STATE(STATE_GET_BSTORE_PAGE);
     return page;
}

static void free_bstore_page(int page)
{
     ENTER_STATE(STATE_FREE_BSTORE_PAGE);

     CILK_ASSERT(page >= 0 && page < bstore_mem / pagesize);

     ENTER_STATE(STATE_WAIT_BS_LOCK);
     Cilk_sema_wait(&bs_lock.lock);
     EXIT_STATE(STATE_WAIT_BS_LOCK);

     bs_flist_array[page] = bs_free_list;
     bs_free_list = page;

     WHEN_CILK_DEBUG(bs_used_pages--);

     Cilk_sema_signal(&bs_lock.lock);
     EXIT_STATE(STATE_FREE_BSTORE_PAGE);
}

/***********************************************************\
 * Frame stack data structures
\***********************************************************/

/* pointer to StealStuffT array */
static volatile StealStuffT *steal_stuff;

/***********************************************************\
 * Page table
\***********************************************************/

/*
 * PAGE_BITS = # of bits that specify a page.  For 32-bit addressing
 * with a page size of 8KB, this gives 19 bits.
 */
#define PAGE_BITS 19

/*
 * LEAF_BITS = # of bits used to index leaf nodes.
 */
#define LEAF_BITS 4

/*
 * NODE_BITS = # of bits used to index internal nodes.
 */
#define NODE_BITS 3

/*
 * LEVELS = # of internal levels of table.
 */
#define NODE_LEVELS 5

/*
 *  We must have PAGE_BITS = LEAF_BITS + NODE_BITS * NODE_LEVELS.
 */

#define MAX_PAGE (1<<PAGE_BITS)
#define LEAF_SIZE (1<<LEAF_BITS)
#define NODE_SIZE (1<<NODE_BITS)

/*
 *  Structure of internal page table nodes.
 */
typedef struct PageTableT {
     LockT lock;
     int refcount;		/* number of incoming pointers */

     volatile void *child[NODE_SIZE];
} PageTableT;

/*
 *  Structure of leaf nodes.
 */
typedef struct {
     LockT lock;
     int refcount;		/* number of incoming pointers */

     int min_valid;		/* minimum valid index */
     volatile void *lower_half;	/* where to go for invalid pages */
     int pages[LEAF_SIZE];
} PageBaseT;

static void setup_page_table(void)
{
     CILK_CHECK(PAGE_BITS == LEAF_BITS + NODE_BITS * NODE_LEVELS,
		("wrong definition of PAGE_BITS\n"));
}

/* 
 * Get base node holding mapping for page <page>-z.  <table> represents
 * pages z through z+<max>-1.
 */
static volatile PageBaseT *get_base(volatile PageTableT *node,
				    int max, int page)
{
     volatile PageTableT *child;
     int step, idx, off;
     int i;

     CILK_ASSERT(node != NULL);
     CILK_ASSERT(page >= 0);
     CILK_ASSERT(page < max);

     if (max == LEAF_SIZE)
	  return (PageBaseT *) node;

     step = max / NODE_SIZE;	/* # of pages for each child */
     idx = page / step;		/* child # of relevant child */
     off = page % step;		/* index in relevant child  */

     child = node->child[idx];

     if (child == NULL) {
	  Cilk_sema_wait(&node->lock);
	  child = node->child[idx];

	  if (child == NULL) {
	       if (step == LEAF_SIZE) {
		    /* allocate new base node and set it up */
		    volatile PageBaseT *base = malloc_shmem(sizeof(PageBaseT));
		    Cilk_sema_init(&base->lock);
		    base->refcount = 1;
		    base->min_valid = 0;
		    base->lower_half = NULL;
		    for (i = 0; i < LEAF_SIZE; i++)
			 base->pages[i] = -1;

		    child = (PageTableT *) base;
	       } else {
		    /* allocate new table node and set it up */
		    child = malloc_shmem(sizeof(PageTableT));
		    Cilk_sema_init(&child->lock);
		    child->refcount = 1;
		    for (i = 0; i < NODE_SIZE; i++)
			 child->child[i] = NULL;
	       }
	       node->child[idx] = child;
	  }
	  Cilk_sema_signal(&node->lock);
     }
     CILK_ASSERT(child != NULL);

     return (get_base(child, step, off));
}

static int get_page_table_entry(volatile PageTableT *table, int page)
{
     volatile PageBaseT *base;
     int idx;
     int bs_page;

     ENTER_STATE(STATE_GET_PAGE_ENTRY);
     CILK_ASSERT(table);
     CILK_ASSERT(table->refcount > 0);
     CILK_ASSERT(page < MAX_PAGE);

     base = get_base(table, MAX_PAGE, page);
     idx = page % LEAF_SIZE;

     while (idx < base->min_valid)
	  base = base->lower_half;

     bs_page = base->pages[idx];
     if (bs_page == -1) {
	  Cilk_sema_wait(&base->lock);
	  bs_page = base->pages[idx];
	  if (bs_page == -1)
	       base->pages[idx] = bs_page = get_bstore_page();
	  Cilk_sema_signal(&base->lock);
     }
     EXIT_STATE(STATE_GET_PAGE_ENTRY);
     return bs_page;
}

/* inc_refcount and dec_refcount work on both PageTableTs and PageBaseTs */
static void inc_refcount(volatile PageTableT *table)
{
     Cilk_sema_wait(&table->lock);
     table->refcount++;
     Cilk_sema_signal(&table->lock);
}

static void dec_refcount(volatile PageTableT *table, int max)
{
     int i;

     Cilk_sema_wait(&table->lock);
     table->refcount--;
     CILK_ASSERT(table->refcount >= 0);
     if (table->refcount == 0) {
	  Cilk_sema_signal(&table->lock);
	  if (max == LEAF_SIZE) {
	       /* free PageBaseT structure */
	       volatile PageBaseT *base = (PageBaseT *) table;
	       for (i = base->min_valid; i < LEAF_SIZE; i++) {
		    if (base->pages[i] != -1)
			 free_bstore_page(base->pages[i]);
	       }
	       if (base->lower_half)
		    dec_refcount(base->lower_half, max);

	       free_shmem(base, sizeof(PageBaseT));
	  } else {
	       /* free PageTableT structure */
	       for (i = 0; i < NODE_SIZE; i++) {
		    if (table->child[i])
			 dec_refcount(table->child[i], max / NODE_SIZE);
	       }
	       free_shmem(table, sizeof(PageTableT));
	  }
	  return;
     }
     Cilk_sema_signal(&table->lock);
}

/* 
 * copy page table, up to and including page z+<page>.  <table>
 * represents pages z through z+<max>-1.
 */
static volatile PageTableT *copy_local(volatile PageTableT *node,
				       int max, int page)
{
     int step, idx, off;
     int i, j;
     volatile PageBaseT *base;
     volatile PageTableT *copy;
     volatile PageBaseT *lower_half;

     CILK_ASSERT(node);
     CILK_ASSERT(page >= 0);
     CILK_ASSERT(page < max - 1);
     CILK_ASSERT(node->refcount > 0);

     if (max == LEAF_SIZE) {
	  /* make new base node */
	  base = malloc_shmem(sizeof(PageBaseT));
	  Cilk_sema_init(&base->lock);
	  base->refcount = 1;
	  for (i = 0; i <= page; i++)
	       base->pages[i] = -2;	/* 
					 * invalid - will never be 
					 * referenced 
					 */
	  for (i = page + 1; i < LEAF_SIZE; i++)
	       base->pages[i] = -1;	/* new entries for new table */
	  base->min_valid = page + 1;

	  /* make sure lower_half chain has decreasig min_valid entries */
	  lower_half = (PageBaseT *) node;
	  while (lower_half->min_valid >= page + 1)
	       lower_half = lower_half->lower_half;
	  CILK_ASSERT(lower_half->min_valid < page + 1);

	  base->lower_half = lower_half;
	  inc_refcount((volatile PageTableT *) lower_half);

	  return ((PageTableT *) base);
     }
     step = max / NODE_SIZE;	/* # of pages for each child */
     idx = page / step;		/* child # of relevant child */
     off = page % step;		/* index in relevant child   */

     copy = malloc_shmem(sizeof(PageTableT));
     Cilk_sema_init(&copy->lock);
     copy->refcount = 1;

     /* make sure children 0 through idx (inclusive) exist */
     for (i = 0; i <= idx; i++) {
	  volatile PageTableT *child = node->child[i];

	  if (child == NULL) {
	       Cilk_sema_wait(&node->lock);
	       child = node->child[i];

	       if (child == NULL) {
		    if (step == LEAF_SIZE) {
			 /* allocate new base node and set it up */
			 base = malloc_shmem(sizeof(PageBaseT));
			 Cilk_sema_init(&base->lock);
			 base->refcount = 1;
			 base->min_valid = 0;
			 base->lower_half = NULL;
			 for (j = 0; j < LEAF_SIZE; j++)
			      base->pages[j] = -1;

			 child = (PageTableT *) base;
		    } else {
			 /* allocate new table node and set it up */
			 child = malloc_shmem(sizeof(PageTableT));
			 Cilk_sema_init(&child->lock);
			 child->refcount = 1;
			 for (j = 0; j < NODE_SIZE; j++)
			      child->child[j] = NULL;
		    }
		    node->child[i] = child;
	       }
	       Cilk_sema_signal(&node->lock);
	  }
     }

     /* copy over low-indexed entries */
     for (i = 0; i < idx; i++) {
	  copy->child[i] = node->child[i];
	  inc_refcount(node->child[i]);
     }

     /* zero out high-indexed entries */
     for (i = idx + 1; i < NODE_SIZE; i++)
	  copy->child[i] = NULL;

     /* if exact, we're done */
     if (off == step - 1) {
	  copy->child[idx] = node->child[idx];
	  inc_refcount(node->child[idx]);
     } else {
	  /* recursively copy table */
	  copy->child[idx] = copy_local(node->child[idx], step, off);
     }
     return copy;
}

/* make a new page table */
static volatile PageTableT *copy_page_table(volatile PageTableT *table,
					    int page)
{
     volatile PageTableT *newtable;
     ENTER_STATE(STATE_COPYING_PAGE_TABLE);
     CILK_ASSERT(page < MAX_PAGE - 1);

     newtable = copy_local(table, MAX_PAGE, page);

     EXIT_STATE(STATE_COPYING_PAGE_TABLE);
     return newtable;
}

/* free page table */
static void free_page_table(volatile PageTableT *table)
{
     ENTER_STATE(STATE_FREEING_PAGE_TABLE);
     CILK_ASSERT(table != NULL);
     dec_refcount(table, MAX_PAGE);
     EXIT_STATE(STATE_FREEING_PAGE_TABLE);
}

static volatile PageTableT *make_new_page_table(void)
{
     volatile PageTableT *table;
     int i;

     table = malloc_shmem(sizeof(PageTableT));
     Cilk_sema_init(&table->lock);
     table->refcount = 1;
     for (i = 0; i < NODE_SIZE; i++)
	  table->child[i] = NULL;

     return table;
}

/*
 * Returns first differing page between tables <t1> and <t2>.  Returns
 * <max> if no pages differ.
 */
static int first_diff_page(volatile PageTableT *t1, volatile PageTableT *t2,
			   int max)
{
     int step, res;
     int i;

     CILK_ASSERT(t1 != NULL);
     CILK_ASSERT(t2 != NULL);
     CILK_ASSERT(t1 != t2);

     if (max == LEAF_SIZE) {
	  for (i = 0; i < LEAF_SIZE; i++) {
	       volatile PageBaseT *b1 = (volatile PageBaseT *) t1;
	       volatile PageBaseT *b2 = (volatile PageBaseT *) t2;
	       int p1, p2;

	       while (i < b1->min_valid)
		    b1 = b1->lower_half;
	       while (i < b2->min_valid)
		    b2 = b2->lower_half;

	       p1 = b1->pages[i];
	       p2 = b2->pages[i];
	       if (p1 == -1 || p2 == -1 || p1 != p2)
		    return i;
	  }
	  return max;
     }
     step = max / NODE_SIZE;	/* # of pages for each child */
     for (i = 0; i < NODE_SIZE; i++) {
	  volatile PageTableT *child1 = t1->child[i];
	  volatile PageTableT *child2 = t2->child[i];

	  if (child1 == NULL || child2 == NULL)
	       return i * step;
	  if (child1 == child2)
	       continue;

	  res = first_diff_page(child1, child2, step);
	  if (res != step)
	       return i * step + res;
     }
     return max;
}

/***********************************************************\
 * Frame stack
\***********************************************************/

static void setup_frame_stack(void)
{
     CILK_CHECK(proc_mem % pagesize == 0,
		("proc_mem not multiple of pagesize!\n"));

     frame_stack = Cilk_map_alloc(proc_mem);
     Cilk_last_mapped_page = -1;

     CILK_CHECK((int) frame_stack % pagesize == 0,
	     ("couldn't align frame_stack to a multiple of pagesize\n"));

     Cilk_allow_faults = 0;
     TRACE(TRACE_SETUP, ("frame stack at %x\n", (unsigned int) frame_stack));
}

/*
 * If fast head-tail check fails, call this slow version.
 */
int Cilk_stolen_check(StackFrameT *frame)
{
     Cilk_sema_wait(&Cilk_ss->stealer_lock);
     if (Cilk_ss->head > frame) {
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);
	  return 1;
     }
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
     return 0;
}

/***********************************************************\
 * page fault catcher
\***********************************************************/

/*
 * Pass this function the fault address for a segmentation violation.
 * Returns 0 if fault was handled successfully, or non-zero if the user
 * program should receive the violation.
 */
int Cilk_signal_handler(void *fault_addr)
{
     signed int fault_page;	/* I emphasise signed -- athena */
     unsigned bs_page;

     ENTER_STATE(STATE_MAPPING_PAGE);
     /* faults can only happen in user code. Otherwise there is a bug */
     CILK_ASSERT(Cilk_allow_faults);

     if (fault_addr < frame_stack ||
	 fault_addr >= frame_stack + proc_mem) {
	  /* fault did not occur in frame stack - we can't handle it */
	  return 1;
     }
     Cilk_allow_faults = 0;

     fault_page = (fault_addr - frame_stack) / pagesize;
     fault_addr = frame_stack + fault_page * pagesize;

     CILK_ASSERT(fault_page >= 0 && fault_page < proc_mem / pagesize);

     if (fault_page > Cilk_last_mapped_page)
	  Cilk_last_mapped_page = fault_page;

     bs_page = get_page_table_entry(Cilk_ss->table, fault_page);

     TRACE(TRACE_SIGNAL,
     ("mapping stack page %d to bstore page %d\n", fault_page, bs_page));

     CILK_ASSERT(bs_page * pagesize + pagesize <= bstore_mem);

     /* map just-faulted page to backing store page */
     Cilk_map_to_bstore(fault_addr, pagesize, bs_page * pagesize);
     CILK_STAT_PAGE_FAULT();

     Cilk_allow_faults = 1;

     EXIT_STATE(STATE_MAPPING_PAGE);
     return 0;
}

/*
 *  Returns a pointer to the ProcInfoT structure describing entry <entry>
 *  of procedure whose signature is p.
 */
static inline ProcInfoT *get_proc_info(ProcInfoT *p, int entry)
{
     CILK_ASSERT(entry > 0);
     return p + entry;
}

/*
 *  Returns a pointer to the slow version for a procedure
 *  whose signature is p.
 */
static void (*get_proc_slow(ProcInfoT *p)) () {
     return p[0].inlet;
}

/***********************************************************\
 * Initialize system
\***********************************************************/

static void global_init(void)
{
     int i;

     /* set weird self value */
     Self = -1;

     /* get OS page size */
     pagesize = Cilk_get_page_size();
     TRACE(TRACE_MISC, ("pagesize = %d\n", pagesize));

     /* setup per-processor and global variables */
     Cilk_map_global_variables();

     /* set up malloc() arena */
     Cilk_init_arena(arenasize);

     /* set up shared memory heap */
     setup_shared_memory();

     /* set up SEGV catcher */
     Cilk_setup_fault_handler();

     /* set up backing store file */
     setup_bstore();

     /* set up virtual memory for frame stack */
     setup_frame_stack();

     /* set up frame stack data structures */
     steal_stuff = Cilk_malloc(Cilk_active_size * sizeof(StealStuffT) +
			       CACHE_LINE - 1);
     steal_stuff = NEXT_CACHE_LINE(steal_stuff);
     for (i = 0; i < Cilk_active_size; i++) {
	  Cilk_sema_init(&steal_stuff[i].stealer_lock);
	  Cilk_sema_wait(&steal_stuff[i].stealer_lock);

	  steal_stuff[i].head = frame_stack;
	  steal_stuff[i].tail = frame_stack;
	  steal_stuff[i].subcomp = NULL;
	  steal_stuff[i].table = NULL;
     }

     /* other shared memory stuff */
     done = 0;

     /* set up page table for each processor */
     setup_page_table();

     /* initialize the statistics gatherer and timer */
     WHEN_CILK_STATS(stats_gathering_init());
     WHEN_CILK_TIMING(time_gathering_init());

     /* initialize library functions */
     lib_function_init();

     /* fork children */
     if (Cilk_begin_parallel()) {
	  /* everybody is done: print stats and exit */
	  print_all_statistics();
#if CILK_DEBUG
	  if (bs_used_pages != 0)
	       CILK_DPRINTF("%d pages of backing store still allocated.\n",
			    bs_used_pages);
	  if (shmem_count != 0)
	       CILK_DPRINTF("%d cache lines of shared memory not freed.\n",
			    shmem_count);
#endif

	  exit(retval);
     }
}

/***********************************************************\
 * Per-processor initialization
\***********************************************************/

static void scheduler_init(void)
{
     /* 
      * make sure random seed is set to something different in each 
      * process 
      */
     rts_srand(Self + 1);

     /* 
      * Set up Cilk_ss pointer.
      */
     Cilk_ss = &steal_stuff[Self];
}

/***********************************************************\
 * Abort code
\***********************************************************/

/*
 * Free child subcomputation structures of <s>.  Each child
 * of <s> is checked to make sure it has returned.  Requires
 * that <s> be locked.  DOES NOT release <s>'s lock before returning.
 */
static void free_children(volatile SubcomputationT *s)
{
     volatile SubcomputationT *child;
     volatile SubcomputationT *next;

     /* splice out child list */
     child = s->down;
     s->down = NULL;

     /* free all children in this list */
     for (; child; child = next) {
	  WHEN_CILK_DEBUG(Cilk_sema_wait(&child->lock));
	  next = child->left;

	  /* All we need to do is deallocate the child. */
	  CILK_ASSERT(child->state == SUBCOMP_RETURNED);
	  CILK_ASSERT(child->down == NULL);
	  WHEN_CILK_DEBUG(child->state = SUBCOMP_INVALID);
	  WHEN_CILK_DEBUG(Cilk_sema_signal(&child->lock));
	  CILK_ASSERT(child->inlet_args == NULL);
	  free_shmem(child, sizeof(SubcomputationT));
     }
}

/*
 * Aborts all children of subcomputation <s>.  Requires
 * that <s> be locked.  DOES NOT release <s>'s lock before returning.
 */
static void abort_children(volatile SubcomputationT *s)
{
     volatile SubcomputationT *child;
     volatile SubcomputationT *next;
     int pn;

     /* splice out child list */
     child = s->down;
     s->down = NULL;

     /* abort all children in this list */
     for (; child; child = next) {
	  TRACE(TRACE_ABORT,
		("abort_children: waiting for child lock, child = %x\n",
		 (int) child));
	  Cilk_sema_wait(&child->lock);
	  TRACE(TRACE_ABORT,
		("abort_children: got child lock\n"));
	  next = child->left;

	  switch (child->state) {
	      case SUBCOMP_RETURNED:
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_RETURNED\n"));
		   /* All we need to do is deallocate the child. */
		   CILK_ASSERT(child->down == NULL);
		   WHEN_CILK_DEBUG(child->state = SUBCOMP_INVALID);
		   Cilk_sema_signal(&child->lock);
		   CILK_ASSERT(child->inlet_args == NULL);
		   free_shmem(child, sizeof(SubcomputationT));
		   break;

	      case SUBCOMP_SYNCD:
		   /* set state to aborting */
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_SYNCD\n"));
		   child->state = SUBCOMP_ABORTING;
		   child->return_loc.size = 0;
		   if (child->return_loc.inlet) {
			if (child->return_loc.argsize > 0)
			     free_shmem(child->inlet_args,
					child->return_loc.argsize);
			WHEN_CILK_DEBUG(child->inlet_args = NULL);
			child->return_loc.inlet = NULL;
		   }
		   child->frame = NULL;
		   /* recursively abort its children */
		   abort_children(child);
		   Cilk_sema_signal(&child->lock);
		   break;

	      case SUBCOMP_RUNNING:
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_RUNNING\n"));
		   /* set state to aborting, set void return value */
		   child->state = SUBCOMP_ABORTING;
		   child->return_loc.size = 0;
		   if (child->return_loc.inlet) {
			if (child->return_loc.argsize > 0)
			     free_shmem(child->inlet_args,
					child->return_loc.argsize);
			WHEN_CILK_DEBUG(child->inlet_args = NULL);
			child->return_loc.inlet = NULL;
		   }
		   child->frame = NULL;
		   /* stop running computation */
		   pn = child->current;
		   CILK_ASSERT(pn != -1);
		   Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
		   CILK_ASSERT(steal_stuff[pn].subcomp == child);
		   steal_stuff[pn].head =
		       (StackFrameT *) (frame_stack + proc_mem);
		   Cilk_sema_signal(&steal_stuff[pn].stealer_lock);

		   /* recursively abort its children */
		   abort_children(child);
		   Cilk_sema_signal(&child->lock);
		   break;

	      case SUBCOMP_ABORTING:
		   Cilk_die("SUBCOMP_ABORTING state in abort_children!\n");
		   break;

	      default:
		   Cilk_die("bad child state!\n");
		   break;
	  }
     }
     TRACE(TRACE_ABORT, ("end of abort_children\n"));
}

static void abort_current_child(volatile SubcomputationT *subcomp)
{
     int pn;
     volatile SubcomputationT *news;
     volatile StackFrameT *frame;

     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING);

     pn = subcomp->current;
     CILK_ASSERT(pn != -1);
     TRACE(TRACE_ABORT,
	   ("abort_current_child: waiting for steal lock\n"));
     Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
     CILK_ASSERT(steal_stuff[pn].subcomp == subcomp);

     /* 
      * we must pretend to steal the top frame, and then immediately sync.
      */
     frame = steal_stuff[pn].head;
     steal_stuff[pn].head = frame + 1;
     Cilk_membar_StoreLoad();
     if (frame + 1 > steal_stuff[pn].tail) {
	  /* no child is currently running - we can just do nothing. */
	  TRACE(TRACE_ABORT,
		("abort_current_child: nothing to steal\n"));
	  steal_stuff[pn].head = frame;
	  Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	  return;
     }
     TRACE(TRACE_ABORT,
	   ("abort_current_child: steal succeeded\n"));

     /* "victim" gets a new subcomputation */
     news = malloc_shmem(sizeof(SubcomputationT));
     Cilk_sema_init(&news->lock);
     news->state = SUBCOMP_ABORTING;
     news->frame = NULL;
     news->join_count = 0;
     news->current = pn;
     news->return_loc.size = 0;
     news->return_loc.index = 0;
     news->return_loc.inlet = NULL;
     WHEN_CILK_DEBUG(news->inlet_args = NULL);
     news->return_subcomp = (SubcomputationT *) subcomp;

     CILK_ASSERT(subcomp->frame == frame);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 1;
		     subcomp->frame->magic = CILK_FRAME_PROVSTEAL_MAGIC;
		     Cilk_allow_faults = 0);
     Cilk_membar_StoreStore();
     subcomp->join_count++;
     Cilk_membar_StoreStore();
     subcomp->current = -1;
     Cilk_membar_StoreStore();
     subcomp->state = SUBCOMP_SYNCD;
     Cilk_membar_StoreStore();

     CILK_ASSERT(steal_stuff[pn].subcomp == subcomp);
     steal_stuff[pn].subcomp = news;

     /* now pretend we abort the newly created subcomputation */
     steal_stuff[pn].head = (StackFrameT *) (frame_stack + proc_mem);

     TRACE(TRACE_ABORT,
	   ("abort_current_child: created subcomp %x\n", (int) news));

     Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
}

void Cilk_abort_from_slow(void)
{
     volatile SubcomputationT *subcomp;

     TRACE(TRACE_ABORT, ("Cilk_abort_from_slow()\n"));

     /* 
      * this code relies on the fact that head >= tail ==> 
      * subcomp is not changed! 
      */
     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);

     subcomp = Cilk_ss->subcomp;
     TRACE(TRACE_ABORT,
       ("Cilk_abort_from_slow: waiting for subcomp lock %x\n", subcomp));
     Cilk_sema_wait(&subcomp->lock);
     TRACE(TRACE_ABORT,
	   ("Cilk_abort_from_slow: waiting for stealer lock\n"));
     Cilk_sema_wait(&Cilk_ss->stealer_lock);
     TRACE(TRACE_ABORT,
	   ("Cilk_abort_from_slow: got stealer lock\n"));
     CILK_ASSERT(subcomp->current == Self);
     CILK_ASSERT(Cilk_ss->subcomp == subcomp);

     abort_children(subcomp);
     Cilk_sema_signal(&subcomp->lock);
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
}

/***********************************************************\
 * Return from a subcomputation
\***********************************************************/

/* 
 * apply_result does whatever <info> says to do to <frame>
 * when it returns.  It deallocates inlet argument structure if
 * necessary.
 */
static void apply_result(ProcInfoT *info, void *inlet_args, StackFrameT *frame)
{
     Cilk_resume = Cilk_abort = 0;
     if (info->inlet) {
	  void (*inlet) (void *, void *, void *) = info->inlet;
	  int index = info->index;
	  int size = info->size;
	  int argsize = info->argsize;

	  /* add just-returned value to beginning of argument list */
	  memcpy((void *) inlet_args, Cilk_result, size);

	  /* run inlet */
	  Cilk_allow_faults = 1;
	  inlet(frame, inlet_args, (char *) frame + index);
	  Cilk_allow_faults = 0;

	  /* free inlet args structure */
	  if (argsize > 0)
	       free_shmem(inlet_args, argsize);
     } else {
	  int index = info->index;
	  int size = info->size;

	  /* copy just-returned value into frame */
	  Cilk_allow_faults = 1;
	  memcpy((char *) frame + index, Cilk_result, size);
	  Cilk_allow_faults = 0;
     }
}

static StackFrameT *suspend_subcomp(volatile SubcomputationT *subcomp)
{	
     /*
      * This code is invoked when a subcomputation suspends without
      * producing a result. This happens when the subcomputation
      * suspended at a sync.
      *
      * subcomp must be locked locked when this function is called.
      * This function releases the lock.
      */
     StackFrameT *return_frame;

     TRACE(TRACE_RETURN_SUBCOMP,
	   ("suspending subcomp %x\n", (int) subcomp));
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("join counter is %d, current = %d return = %x \n",
	    subcomp->join_count,
	    subcomp->current,
	    subcomp->return_subcomp));

     CILK_ASSERT(subcomp->frame != NULL);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 1);
     CILK_ASSERT((subcomp->frame)->magic == CILK_FRAME_NONSTEALABLE_MAGIC);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 0);

     if (subcomp->join_count == 0) {
	  /* 
	   * the join_count was non-zero when the subcomputation
	   * suspended, but it is zero now.  We need to restart
	   * this subcomputation.  We will treat this condition as
	   * a provably good steal.
	   */
	  TRACE(TRACE_RETURN_SUBCOMP,
		("restarting subcomp %x\n", (int) subcomp));

	  /* splice subcomputation back into steal stuff */
	  Cilk_ss->subcomp = subcomp;
	  subcomp->current = Self;
	  
	  return_frame = subcomp->frame;
	  Cilk_sema_signal(&subcomp->lock);
	  return return_frame;
     }
	
     /* 
      * check that the frame is non stealable, but make it stealable,
      * for use of provably good steals.
      */
     CILK_ASSERT((subcomp->frame)->magic ==
		 CILK_FRAME_NONSTEALABLE_MAGIC);
     WHEN_CILK_DEBUG((subcomp->frame)->magic = CILK_FRAME_PROVSTEAL_MAGIC);
     
     /* set subcomputation state */
     if (subcomp->state != SUBCOMP_ABORTING)
	  subcomp->state = SUBCOMP_SYNCD;
     
     Cilk_sema_signal(&subcomp->lock);
     return NULL;
}

static void returnit_from_main(volatile SubcomputationT *subcomp)
{
     /* 
      * the suspended subcomputation produces a value returned by
      * cilk_main. subcomp is locked at this point.
      */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("return of cilk_main\n"));
     retval = *((int *) Cilk_result);
     WHEN_CILK_CP(Cilk_critical_path = Cilk_cp_hack);

     done = 1;

     /* finish with returning subcomputation */
     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING);
     free_children(subcomp);
     WHEN_CILK_DEBUG(subcomp->state = SUBCOMP_INVALID);
     CILK_ASSERT(subcomp->inlet_args == NULL);
     Cilk_sema_signal(&subcomp->lock);
     free_shmem(subcomp, sizeof(SubcomputationT));
     return;
}

/*
 * Sends the return value of the finished subcomputation to wherever it is
 * supposed to go.  Returns a frame to run obtained by a provably good steal,
 * if any.
 *
 * This code is messy, sorry about that.
 */
static StackFrameT *returnit(void)
{
     volatile SubcomputationT *subcomp;
     volatile SubcomputationT *return_subcomp;
     StackFrameT *return_frame;
     ProcInfoT return_loc;
     int pn;
     void *inlet_args;

     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("returnit(): head = %x tail = %x\n", (int) Cilk_ss->head,
	    (int) Cilk_ss->tail));
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("returnit(): Cilk_result_valid = %d, result=%d\n",
	    Cilk_result_valid, *(int *) Cilk_result));

     /* 
      * Get current subcomputation.  Make sure we get subcomputation
      * lock before stealer lock.  This code relies on the fact
      * that head >= tail ==> subcomp is not changed!
      */
     subcomp = Cilk_ss->subcomp;

     /* get rid of this and the lock in steal()/main() ? */
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
     Cilk_sema_wait(&subcomp->lock);
     Cilk_sema_wait(&Cilk_ss->stealer_lock);

     CILK_ASSERT(subcomp->current == Self);
     CILK_ASSERT(Cilk_ss->subcomp == subcomp);
     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING ||
		 subcomp->state == SUBCOMP_ABORTING);

     /* splice current subcomp out of steal stuff */
     Cilk_ss->subcomp = NULL;
     subcomp->current = -1;

     /* 
      * Treat an abort as a return with a value if the join count is zero
      * and a return without a value if the join count is non-zero.
      */
     if (subcomp->state == SUBCOMP_ABORTING) {
	  CILK_ASSERT(subcomp->return_loc.size == 0);
	  CILK_ASSERT(subcomp->return_loc.inlet == NULL);
	  if (subcomp->join_count > 0) {
	       Cilk_sema_signal(&subcomp->lock);
	       return NULL;
	  }
	  Cilk_result_valid = 1;
     }

     if (!Cilk_result_valid)
	  return suspend_subcomp(subcomp);

     /* 
      * Otherwise, the frame produces a result. We must send the result to
      * the correct location.
      */
   produces_result:

     /* some debugging stuff */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("finishing subcomp %x\n", (int) subcomp));
     if (subcomp->return_loc.size == 4)
	  TRACE(TRACE_RETURN_SUBCOMP,
		("returning %d\n", *((int *) Cilk_result)));
     CILK_ASSERT(subcomp->join_count == 0);

     /* get subcomputation to return to */
     return_subcomp = subcomp->return_subcomp;

     if (return_subcomp == NULL) {
	  /* return value of cilk_main: we're done! */
	  returnit_from_main(subcomp);
	  return NULL;
     }

     CILK_ASSERT(return_subcomp->current != Self);

     /* get place to put return value */
     return_loc = subcomp->return_loc;
     inlet_args = (void *) subcomp->inlet_args;
     WHEN_CILK_DEBUG(subcomp->inlet_args = NULL);

     /* finish with returning subcomputation */
     if (subcomp->state == SUBCOMP_ABORTING) {
	  WHEN_CILK_DEBUG(subcomp->state = SUBCOMP_INVALID);
	  CILK_ASSERT(subcomp->down == NULL);
	  Cilk_sema_signal(&subcomp->lock);
	  CILK_ASSERT(inlet_args == NULL);
	  free_shmem(subcomp, sizeof(SubcomputationT));
     } else {
	  subcomp->state = SUBCOMP_RETURNED;
	  free_children(subcomp);
	  Cilk_sema_signal(&subcomp->lock);
     }

     /* get lock on return frame */
   try_again:
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("trying to acquire subcomp lock, return_subcomp = %x\n",
	    return_subcomp));
     Cilk_sema_wait(&return_subcomp->lock);
     return_frame = return_subcomp->frame;

     /*
      * take returned value and apply it to the returned-to frame. 
      * Only do it if the returned-to frame isn't aborted.
      */
     if (return_subcomp->state != SUBCOMP_ABORTING) {
	  CILK_ASSERT(return_frame != NULL);
	  pn = return_subcomp->current;
	  if (pn == -1) {
	       /* 
	        * just run it - we don't need to worry about local
	        * worker 
	        */
	       apply_result(&return_loc, inlet_args, return_frame);
	  } else {
	       /* 
	        * Some processor is working on this subcomputation.
	        * Get a lock on that processor, make sure he's still
	        * the worker, get a lock on the return frame, and then
	        * apply the result.
	        */

	       /* lock owning processor */
	       Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
	       CILK_ASSERT(steal_stuff[pn].subcomp == return_subcomp);

	       /* get lock on frame */
	       CILK_ASSERT(return_frame == steal_stuff[pn].head);
	       steal_stuff[pn].head = return_frame + 1;
	       Cilk_membar_StoreLoad();
	       if (return_frame + 1 > steal_stuff[pn].tail) {
		    steal_stuff[pn].head = return_frame;
		    Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
		    Cilk_sema_signal(&return_subcomp->lock);
		    for (pn = 0; pn < 1000; pn++);	/* some delay */
		    goto try_again;
	       }
	       apply_result(&return_loc, inlet_args, return_frame);

	       /* unlock frame */
	       steal_stuff[pn].head = return_frame;
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	  }
	  
	  /* the inlet may have aborted --- check for that */
	  if (Cilk_abort) {
	       TRACE(TRACE_ABORT,
		     ("inlet set abort flag\n"));
	       if (return_subcomp->state == SUBCOMP_RUNNING)
		    abort_current_child(return_subcomp);
	       abort_children(return_subcomp);
	  }
     } else {
	  /* the subcomputation is aborting */
	  if (return_loc.inlet != NULL && return_loc.argsize > 0)
	       free_shmem(inlet_args, return_loc.argsize);
     }

     /* 
      * decrement join counter of returned-to subcomputation, and 
      * do a provably good steal if necessary.
      */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("enabling subcomp %x, join = %d (before enabling), "
	    "current = %d\n", (int) return_subcomp,
	    return_subcomp->join_count, return_subcomp->current));
     CILK_ASSERT(return_subcomp->state == SUBCOMP_RUNNING ||
		 return_subcomp->state == SUBCOMP_SYNCD ||
		 return_subcomp->state == SUBCOMP_ABORTING);

     /* 
      * update critical path 
      */
     WHEN_CILK_CP(if (Cilk_cp_hack > return_subcomp->cp)
		  return_subcomp->cp = Cilk_cp_hack);

     /* 
      * make sure to commit all changes to memory before enabling
      * a remote subcomputation.
      */
     Cilk_membar_StoreStore();

     return_subcomp->join_count--;
     CILK_ASSERT(return_subcomp->join_count >= 0);

     if (return_subcomp->current == -1 &&
	 (return_subcomp->join_count == 0 || Cilk_resume)) {
	  /* that is, if nobody is working on a ready subcomputation */

	  TRACE(TRACE_RETURN_SUBCOMP,
		("adopting subcomp %x\n", (int) return_subcomp));

	  if (return_subcomp->state == SUBCOMP_ABORTING) {
	       /* 
	        * don't restart the subcomputation.  Just pretend it
	        * immediately returns a value.
	        */
	       TRACE(TRACE_ABORT,
		     ("the adopted subcomputation is aborting\n"));
	       subcomp = return_subcomp;
	       goto produces_result;
	  }
	  return_subcomp->current = Self;
	  return_subcomp->state = SUBCOMP_RUNNING;
	  Cilk_ss->subcomp = return_subcomp;
	  WHEN_CILK_DEBUG(Cilk_allow_faults = 1);
	  CILK_ASSERT(return_frame->magic == CILK_FRAME_PROVSTEAL_MAGIC);
	  WHEN_CILK_DEBUG(Cilk_allow_faults = 0);
	  Cilk_sema_signal(&return_subcomp->lock);

	  return return_frame;
     }
     Cilk_sema_signal(&return_subcomp->lock);
     return NULL;
}

/***********************************************************\
 * Work stealing
\***********************************************************/

static inline int min(int x, int y)
{
     return x < y ? x : y;
}

static inline int max(int x, int y)
{
     return x > y ? x : y;
}

static void runit(volatile StackFrameT *frame)
{
     /* 
      * execute the frame, and keep doing provably good steals while we can 
      */
     while(frame) {
	  Cilk_ss->head = frame;
	  Cilk_ss->tail = frame;
	  Cilk_result_valid = 0;
	  TRACE(TRACE_STEAL, ("running at entry %d\n", frame->entry));
	  
	  Cilk_allow_faults = 1;

	  WHEN_CILK_DEBUG(frame->magic = CILK_FRAME_NONSTEALABLE_MAGIC);

	  /* allow stealing */
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);
	  
	  ENTER_STATE(STATE_WORKING);
	  (get_proc_slow(frame->sig)) (frame);
	  EXIT_STATE(STATE_WORKING);
	  
	  /* disallow stealing */
	  Cilk_sema_wait(&Cilk_ss->stealer_lock);

	  Cilk_allow_faults = 0;

	  frame = returnit();
	  TRACE(TRACE_RETURN_SUBCOMP,
		("steal: returnit() = %x\n", (int) frame));
     }
}

static void steal(void)
{
     int pn;
     volatile StackFrameT *frame;
     volatile StackFrameT *next_frame;
     int page, page1, page2;
     volatile SubcomputationT *s, *news;
     ProcInfoT *info;
     int bs_page;
     void *page_addr;
     volatile PageTableT *table, *old_table;

     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);
     ENTER_STATE(STATE_STEALING);
     TRACE(TRACE_STEAL, ("stealing\n"));

     CILK_ASSERT(Cilk_ss->subcomp == NULL);

     /* Keep trying until we successfully steal. */
     while (!done) {
	  /* pick a random processor, not ourself */
	  pn = ((unsigned) rts_rand()) % Cilk_active_size;
	  if (pn == Self)
	       continue;

	  TRACE(TRACE_STEAL, ("trying to steal from %d\n", pn));
	  CILK_STAT_STEAL_ATTEMPT();
	  /* try to get exclusive stealer lock */
	  if (!Cilk_sema_try(&steal_stuff[pn].stealer_lock)) {
	       TRACE(TRACE_STEAL, ("lock is unavailable\n", pn));
	       continue;
	  }
	  TRACE(TRACE_STEAL, ("got lock on %d\n", pn));

	  TRACE(TRACE_STEAL, ("trying to get current subcomp on %d\n", pn));
	  s = steal_stuff[pn].subcomp;
	  CILK_ASSERT(s != NULL);
	  if (!Cilk_sema_try(&s->lock)) {
	       TRACE(TRACE_STEAL, ("current subcomp is unavailable on %d\n", pn));
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       continue;
	  }
	  CILK_ASSERT(s->current == pn);
	  if (s->state == SUBCOMP_ABORTING) {
	       TRACE(TRACE_STEAL, ("current subcomp is aborting on %d\n", pn));
	       Cilk_sema_signal(&s->lock);
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       continue;
	  }
	  CILK_ASSERT(s->state == SUBCOMP_RUNNING);
	  TRACE(TRACE_STEAL, ("got current subcomp on %d\n", pn));

	  /* 
	   * This is a Dekker-like protocol for mutual exclusion, using
	   * shared memory.
	   *
	   * Thief:
	   *   lock
	   *   H++
	   *   --MEMBAR-- ensure H write occurs before T is read.
	   *   If (H>T)
	   *     H--
	   *     unlock
	   *     give up
	   *   steal
	   *   unlock
	   *
	   * Victim
	   *   T--
	   *   --MEMBAR-- ensure T write occurs before H is read.
	   *   if (H>T)
	   *      lock
	   *      if (H>T)
	   *        unlock
	   *        give up
	   *      unlock
	   *
	   */

	  frame = steal_stuff[pn].head;

	  /* 
	   * Increment by sizeof(StackFrameT) for now.  We must increment
	   * by frame->size eventually.
	   */
	  steal_stuff[pn].head = frame + 1;
	  Cilk_membar_StoreLoad();

	  if (frame + 1 > steal_stuff[pn].tail) {
	       steal_stuff[pn].head = frame;

	       /* nothing to steal -- give up */
	       TRACE(TRACE_STEAL, ("nothing to steal on %d\n", pn));
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       Cilk_sema_signal(&s->lock);
	       continue;
	  }
	  TRACE(TRACE_STEAL, ("stealing from %d\n", pn));
	  CILK_STAT_STEAL();

	  /* ensure we will read fresh values from memory */
	  Cilk_membar_LoadLoad();

	  /* 
	   * Get page table of victim and mark it so it won't be deallocated.
	   */
	  table = steal_stuff[pn].table;
	  inc_refcount(table);

	  /* 
	   * map pages containing frame header
	   */
	  page1 = ((char *) frame - (char *) frame_stack) / pagesize;
	  page2 = ((char *) frame + sizeof(StackFrameT) - 1
		   - (char *) frame_stack) / pagesize;
	  for (page = page1; page <= page2; page++) {
	       bs_page = get_page_table_entry(table, page);
	       page_addr = frame_stack + page * pagesize;
	       Cilk_map_to_bstore(page_addr, pagesize, bs_page * pagesize);
	       CILK_STAT_PAGE_FAULT();
	  }

	  /* now we can assert */
	  CILK_ASSERT(frame->magic == CILK_FRAME_STEALABLE_MAGIC);
	  WHEN_CILK_DEBUG(frame->magic = CILK_FRAME_STOLEN_MAGIC);

	  /* 
	   * now that the memory map is set up, we can
	   * find out the size of the frame we stole, and hence where
	   * the beginning of the next frame is.
	   */
	  next_frame = (StackFrameT *) ((char *) frame + frame->size);

	  /* Set head of victim's queue correctly */
	  steal_stuff[pn].head = next_frame;

	  /* 
	   * We now proceed to arrange the subcomputations
	   * to keep the linkage intact. 
	   */
	  TRACE(TRACE_STEAL,
		("stolen frame has entry %d\n", frame->entry));

	  /* victim gets a new subcomputation */
	  news = malloc_shmem(sizeof(SubcomputationT));
	  Cilk_sema_init(&news->lock);
	  news->state = SUBCOMP_RUNNING;
	  news->frame = (StackFrameT *) next_frame;
	  news->join_count = 0;
	  news->current = pn;
	  WHEN_CILK_TIMING(news->cp = 0);

	  info = get_proc_info(frame->sig, frame->entry);
	  news->return_subcomp = (SubcomputationT *) s;
	  news->return_loc = *info;
	  if (news->return_loc.index == -1)
	       news->return_loc.index = frame->index;
	  if (info->inlet != NULL && info->argsize > 0) {
	       /* get arguments of inlet out of frame and into subcomp. */
	       volatile void *inlet_args = malloc_shmem(info->argsize);
	       Cilk_allow_faults = 1;
	       memcpy((void *) inlet_args, (char *) frame + info->argindex,
		      info->argsize);
	       Cilk_allow_faults = 0;
	       news->inlet_args = inlet_args;
	  }
	  TRACE(TRACE_STEAL,
		("new subcomp %x, return to %x, index %d, size %d\n",
		 (int) news, (int) news->return_subcomp,
		 news->return_loc.index, news->return_loc.size));

	  /* stealer inherits victim's subcomputation */
	  Cilk_ss->subcomp = s;
	  s->current = Self;
	  s->join_count++;
	  TRACE(TRACE_STEAL,
	  ("stealing subcomp: %x, join = %d\n", (int) s, s->join_count));

	  /* abort linkage */
	  news->down = NULL;
	  news->left = s->down;
	  s->down = news;

	  /* make news the current subcomputation of the victim */
	  steal_stuff[pn].subcomp = news;

	  /* tell victim we're done stealing */
	  Cilk_sema_signal(&s->lock);
	  Cilk_sema_signal(&steal_stuff[pn].stealer_lock);

	  /* 
	   * Make page faults happen on all pages which:
	   * 1) differ from our old table (>= first_diff_page),
	   * 2) may be mapped (<= Cilk_last_mapped_page), and
	   * 3) we haven't mapped correctly yet (page1 to page2, inclusive).
	   */
	  old_table = Cilk_ss->table;
	  if (old_table != NULL) {
	       page = first_diff_page(old_table, table, MAX_PAGE);
#if 0
	       /* 
	        * Note that this test has side-effects: it fills in
	        * entries in the page table that might not otherwise
	        * be filled.
	        */
	       {
		    int i;
		    for (i = 0; i < page; i++)
			 CILK_ASSERT(get_page_table_entry(old_table, i) ==
				     get_page_table_entry(table, i));
		    CILK_ASSERT(get_page_table_entry(old_table, i) !=
				get_page_table_entry(table, i));
	       }
#endif
	  } else {
	       page = 0;
	       CILK_ASSERT(Cilk_last_mapped_page == -1);
	  }

	  /* 
	   * unmap the portion of stack from page to Cilk_last_mapped_page,
	   * excluding the chunk from page1 to page2 we just mapped in.
	   */
	  {
	       int start1 = page;
	       int end1 = min(page1 - 1, Cilk_last_mapped_page);
	       int start2 = page2 + 1;
	       int end2 = Cilk_last_mapped_page;

	       if (start1 <= end1)
		    Cilk_map_fault(frame_stack + start1 * pagesize,
				   (end1 - start1 + 1) * pagesize);
	       if (start2 <= end2)
		    Cilk_map_fault(frame_stack + start2 * pagesize,
				   (end2 - start2 + 1) * pagesize);
	       Cilk_last_mapped_page = page2;
	  }

	  /* 
	   * we now copy the page table of the victim.
	   */
	  page = ((char *) next_frame - 1 - (char *) frame_stack) / pagesize;
	  CILK_ASSERT(page >= page2);
	  Cilk_ss->table = copy_page_table(table, page);

	  /* 
	   * Undo mark to allow victim's table to be deallocated.
	   */
	  free_page_table(table);

	  /* 
	   * Undo mark to allow old page table to be deallocated.
	   */
	  if (old_table != NULL)
	       free_page_table(old_table);

	  /* 
	   * if stolen frame's end is not page-aligned, add some size to
	   * it to make it page-aligned. 
	   */
	  next_frame = (StackFrameT *) (((unsigned) next_frame
				  + pagesize - 1) / pagesize * pagesize);
	  frame->size = (char *) next_frame - (char *) frame;

	  /* run stolen frame */
	  EXIT_STATE(STATE_STEALING);
	  runit(frame);
	  return;
     }
     EXIT_STATE(STATE_STEALING);
}

/***********************************************************\
 * Command line parsing
\***********************************************************/

enum {
     NONE, NPROC, STATS, NO_STATS, HELP, DEBUGOPT, ARENA,
     STACK, BSTORE_SIZE, END_OPTIONS
};

static struct options {
     char *string;
     int option;
} optarray[] = {

     {
	  "-", END_OPTIONS	/* -- forces end to cilk options */
     },
     {
	  "nproc", NPROC
     },
     {
	  "stats", STATS
     },
     {
	  "no-stats", NO_STATS
     },
     {
	  "help", HELP
     },
     {
	  "debug", DEBUGOPT
     },
     {
	  "heap", ARENA
     },
     {
	  "arena", ARENA
     },
     {
	  "stack", STACK
     },
     {
	  "bsize", BSTORE_SIZE
     },
     {
	  (char *) 0, NONE
     }
};

static void print_help(void)
{
     printf("Cilk options:\n");
     printf("  -nproc <number of processors>\n");
     printf("  -stats <statistics level>\n");
     printf("  -no-stats\n");
     printf("  -debug <trace mode>\n");
     printf("  -bfile <backing store file>\n");
     printf("  -arena <size in MB> | -heap <size in MB>\n");
     printf("  -stack <size in MB>\n");
     printf("  -bsize <size in MB>\n");
     printf("  -help\n");
     printf("Use `--' to force Cilk option parsing to stop.\n");
}

/* look for a given string in the option table */
static int parse_option(char *s)
{
     struct options *p;

     for (p = optarray; p->string; ++p)
	  if (strcmp(s, p->string) == 0)
	       break;

     return p->option;
}

static int parse_command_line(int *argc, char *argv[])
{
     int i, j;
     for (i = 1; i < *argc; ++i) {
	  if (*argv[i] != '-')
	       goto arch_specific;
	  switch (parse_option(argv[i] + 1)) {
	      case NPROC:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-nproc requires a number\n"));
		   Cilk_active_size = atoi(argv[i]);
		   if (Cilk_active_size == 0)
			Cilk_active_size = Cilk_partition_size();
		   CILK_CHECK(Cilk_active_size > 0,
			      ("non-positive number of processors\n"));
		   break;
	      case STATS:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-status requires a number\n"));
		   CilkStatLevel = atoi(argv[i]);
		   break;
	      case NO_STATS:
		   CilkStatLevel = 0;
		   break;
	      case DEBUGOPT:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-debug requires a number\n"));
		   Cilk_trace_mode = strtol(argv[i], (char **) 0, 0);
		   break;
	      case HELP:
		   print_help();
		   return 1;
		   break;
	      case ARENA:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires an integer\n", argv[i - 1]));
		   arenasize = atoi(argv[i]) * MEGABYTE;
		   break;
	      case STACK:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires an integer\n", argv[i - 1]));
		   proc_mem = atoi(argv[i]) * MEGABYTE;
		   break;
	      case BSTORE_SIZE:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires an integer\n", argv[i - 1]));
		   bstore_mem = atoi(argv[i]) * MEGABYTE;
		   break;
	      case END_OPTIONS:
		   ++i;
		   goto stop;
	      default:
		   goto arch_specific;
	  }
     }

     /* let architecture-specific code get a crack at the command line */
   arch_specific:
     i = Cilk_arch_specific_options(i, *argc, argv);

     /* remove Cilk command-line arguments from argv */
   stop:
     for (j = i; j <= *argc; j++)
	  argv[j - (i - 1)] = argv[j];
     *argc -= i - 1;

     return 0;
}

/***********************************************************\
 * Main routine
\***********************************************************/

/* prototype for cilk_main needed here */
int cilk_main(int argc, char *argv[]);

/* Hack: main is defined to be cilk_main; undef it */
#undef main

int main(int argc, char *argv[])
{
     /* start elapsed time measurement */
     start_time = Cilk_get_wall_time();

     /* parse command-line arguments */
     if (parse_command_line(&argc, argv))
	  return 0;

     /* initialize global part of system */
     global_init();

     TRACE(TRACE_MISC, ("running\n"));

     /* start gathering timing statistics */
     ENTER_STATE(STATE_TOTAL);

     /* initialize per-processor parts of system */
     scheduler_init();

     /* start up cilk_main on processor 0 */
     if (Self == 0) {
	  int res;
	  StackFrameT *frame;
	  volatile SubcomputationT *s = malloc_shmem(sizeof(SubcomputationT));
	  s->state = SUBCOMP_RUNNING;
	  s->return_subcomp = NULL;
	  Cilk_sema_init(&s->lock);
	  s->frame = frame_stack;
	  s->current = 0;
	  s->join_count = 0;
	  s->down = s->left = NULL;
	  Cilk_ss->subcomp = s;

	  CILK_ASSERT(Cilk_ss->table == NULL);
	  Cilk_ss->table = make_new_page_table();

	  Cilk_result_valid = 0;

	  TRACE(TRACE_MAIN, ("invoking cilk_main()\n"));

	  /* start with a 0 critical path */
	  WHEN_CILK_CP(Cilk_cp_hack = 0);

	  Cilk_allow_faults = 1;

	  /* allow stealing */
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);

	  ENTER_STATE(STATE_WORKING);
	  res = cilk_main(argc, argv);
	  EXIT_STATE(STATE_WORKING);

	  /* disallow stealing */
	  Cilk_sema_wait(&Cilk_ss->stealer_lock);

	  Cilk_allow_faults = 0;

	  TRACE(TRACE_MAIN, ("cilk_main returned\n"));

	  if (!Cilk_result_valid) {
	       /* cilk_main wasn't stolen - set result values */
	       TRACE(TRACE_MAIN,
		     ("main was not stolen\n"));
	       CILK_ASSERT(Cilk_ss->subcomp->return_subcomp == NULL);
	       *((int *) Cilk_result) = res;
	       Cilk_result_valid = 1;
	  } else {
	       TRACE(TRACE_MAIN,
		     ("main was stolen\n"));
	  }

	  TRACE(TRACE_MAIN, ("main: invoking returnit()\n"));
	  frame = returnit();
	  TRACE(TRACE_RETURN_SUBCOMP,
		("main: returnit() = %x\n", (int) frame));

	  /* keep doing provably good steals while we can */
	  TRACE(TRACE_MAIN, ("main: doing provably good steals\n"));
	  runit(frame);
     }

     TRACE(TRACE_MAIN, ("main: stealing normally\n"));
     /* keep stealing until we're done */
     while (!done)
	  steal();

     /* free any leftover page table */
     if (Cilk_ss->table)
	  free_page_table(Cilk_ss->table);

     TRACE(TRACE_MAIN, ("main: done!\n"));

     /* stop gathering time statistics */
     EXIT_STATE(STATE_TOTAL);

     /* kill child processes */
     Cilk_end_parallel();

     /* never reached! */
     CILK_ASSERT(0 /* can't happen */ );
     return 0;
}

