/*
 * $Id: cilk.h,v 1.59 1996/06/07 17:55:50 athena Exp $
 */
/*
 * $Log: cilk.h,v $
 * Revision 1.59  1996/06/07 17:55:50  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.58  1996/06/06 14:37:43  athena
 * cilk_main is now main
 *
 * Revision 1.57  1996/06/06 00:18:11  athena
 * Added printout of total execution time.
 *
 * Revision 1.56  1996/06/05 15:37:47  athena
 * Added cilk-lib.h, and updated stuff
 *
 * Revision 1.55  1996/06/03 21:20:48  athena
 * Minor cleanup
 *
 * Revision 1.54  1996/05/30 21:53:50  randall
 * changed prototype for Cilk_srand
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_H
#define _CILK_H

#include <cilk-conf.h>

/***********************************************************\
 * Architecture-specific definitions and functions
\***********************************************************/

#include CILK_ARCH_H

/***********************************************************\
 * Types used by cilk2c output
\***********************************************************/

/*
 * ProcInfoT fields:
 *          |       NORMAL           |               INLET
 * ---------+------------------------+---------------------------------------
 *     size | size of return value   | size of returned (by SPAWN) value
 *    index | index to put ret. val. | index to put ret. (by INLET_CALL) val.
 *    inlet | 0                      | pointer to inlet function
 *  argsize | ???                    | size of inlet's arguments in frame
 * argindex | ???                    | index of inlet's arguments in frame
 *
 * Hack: the field "inlet" for entry 0 contains the address of 
 * the slow function.
 */
typedef struct {
     int size;
     int index;
     void (*inlet) ();
     int argsize;
     int argindex;
} ProcInfoT;

typedef struct StackFrameT {
     int size;
     int entry;
     int index;			/* 
				 * index outstanding spawn will return to, 
				 * valid only if entry does not specify
				 * it. 
				 */
     ProcInfoT *sig;
     WHEN_CILK_TIMING(TimeT mycp;)
     WHEN_CILK_TIMING(TimeT cp;)
     WHEN_CILK_DEBUG(volatile int magic;)
} StackFrameT;

typedef struct SubcomputationT {
     LockT lock;
     int state;

     /* subcomputation to return value to */
     struct SubcomputationT *return_subcomp;

     /* place in frame to put return value */
     ProcInfoT return_loc;

     StackFrameT *frame;	/* base frame of subcomputation */

     /* 
      * Number of processor currently executing subcomp.  
      * -1 if no one is executing it.
      */
     volatile int current;

     /* number of outstanding children (excluding current, if any) */
     volatile int join_count;
     volatile void *inlet_args;	/* arguments for inlet to run */

     /* list of child subcomputations */
     volatile struct SubcomputationT *down;

     /* left sibling computation */
     volatile struct SubcomputationT *left;

     /* critical path of the subcomputation */
     WHEN_CILK_TIMING(volatile TimeT cp;)
} SubcomputationT;

/*
 *  Possible states of a subcomputation.
 *   SUBCOMP_INVALID   can't ever be this.
 *   SUBCOMP_RUNNING   currently running on some processor
 *   SUBCOMP_SYNCD     not running on any processor, waiting for children
 *   SUBCOMP_ABORTING  marked as aborted, waiting for currently running
 *                     computation to complete.
 *   SUBCOMP_RETURNED  returned to parent, waiting to be deallocated.
 */
#define SUBCOMP_INVALID		0
#define SUBCOMP_RUNNING		1
#define SUBCOMP_SYNCD		2
#define SUBCOMP_ABORTING	3
#define SUBCOMP_RETURNED	4

typedef struct StealStuffT {
     /* lock to ensure one stealer at a time */
     CACHE_LINE_ALIGNED(LockT stealer_lock);

     volatile StackFrameT *head;	/* head of frame queue */
     volatile StackFrameT *tail;	/* tail of frame queue */

     /* subcomputation currently running */
     volatile SubcomputationT *subcomp;

     /* page table of running computation */
     volatile struct PageTableT *table;
} StealStuffT;

/***********************************************************\
 * Variables defined by cilk2c output
\***********************************************************/

/* None --- isn't it great? :-) */

/***********************************************************\
 * Variables used by cilk2c output
\***********************************************************/

/* per-processor variables used by cilk2c output */
extern int Cilk_result_valid;
extern int Cilk_resume;
extern int Cilk_abort;

/* Hack for critical path measurements */
WHEN_CILK_TIMING(extern TimeT Cilk_cp_hack;)

/* global variables used by cilk2c output */
extern char Cilk_result[];	/* TODO */

/* functions used by cilk2c output */
extern int Cilk_stolen_check(StackFrameT *frame);
extern void Cilk_abort_from_slow(void);

/* define a static message. gcc would duplicate the string otherwise */
extern const char *Cilk_assertion_failed_msg;

/* errors and debugging */
extern long Cilk_trace_mode;

/***********************************************************\
 * Internal rts functions
\***********************************************************/

extern void Cilk_init_arena(size_t size);
extern int Cilk_signal_handler(void *fault_addr);

extern char Cilk_rts_data_begin[];
extern char Cilk_rts_data_mid[];

/***********************************************************\
 * Hacks
\***********************************************************/
#define main cilk_main

#endif				/* _CILK_H */
